from typing import Annotated

import numpy as np
from numpy.typing import NDArray

# Basic numerical types
type Real = float | np.float64
type Complex = complex | np.complex128

# General vector and matrix types
type FloatVector = NDArray[np.float64]  # Shape: (n,)
type ComplexVector = NDArray[np.complex128]  # Shape: (n,)
type Float2D = NDArray[np.float64]  # Shape: (n, m)
type Complex2D = NDArray[np.complex128]  # Shape: (n, m)

type AtomicCoordinates = NDArray[np.float64]  # Shape: (n_atoms, 3)
type AtomicNumbers = NDArray[np.int32]  # Shape: (n_atoms,)

# Type for atomic symbols with validation
type AtomicSymbol = Annotated[str, "Valid chemical element symbol"]

# Energy types with units (keeping these as they're fundamental)
type Energy = Annotated[float, "Energy in Hartree"]
type EnergyEV = Annotated[float, "Energy in electron volts"]



from qiskit.circuit.library import PauliEvolutionGate
# Trotter-Suzuki implementation for decomposition of exponentials
# of matrices
from qiskit.synthesis import SuzukiTrotter
from qiskit.quantum_info import SparsePauliOp
from qiskit import QuantumCircuit, QuantumRegister
import numpy as np
from itertools import groupby
import re



# -----------------------------------------------------------------
# Hamiltonian Functions
# -----------------------------------------------------------------
def get_hamiltonian_n_site_terms(n, coeff, n_qubits):
    '''
        Assembles each term in the Hamiltonian based on their Pauli string
        representation and multiplying by the respective coefficient.
    '''
    XX_coeff = coeff[0]
    YY_coeff = coeff[1]
    ZZ_coeff = coeff[2]
    Z_coeff = coeff[3]

    XX_term = SparsePauliOp(("I" * n + "XX" + "I" * (n_qubits - 2 - n)))
    XX_term *= XX_coeff
    YY_term = SparsePauliOp(("I" * n + "YY" + "I" * (n_qubits - 2 - n)))
    YY_term *= YY_coeff
    ZZ_term = SparsePauliOp(("I" * n + "ZZ" + "I" * (n_qubits - 2 - n)))
    ZZ_term *= ZZ_coeff
    Z_term = SparsePauliOp(("I" * n + "Z" + "I" * (n_qubits - 1 - n)))
    Z_term *= Z_coeff

    return (XX_term + YY_term + ZZ_term + Z_term)


def get_heisenberg_hamiltonian(n_qubits, coeff=None):
    r'''
        Takes an integer number corresponding to number of spins/qubits
        and a list of sublists containing the necessary coefficients
        to assemble the complete Hamiltonian:
        $$
            H = \sum _i ^N h_z Z_i
                + \sum _i ^{N-1} (h_xx X_i X_{i+1}
                                  + h_yy Y_i Y_{i+1}
                                  + h_zz Z_i Z_{i+1}
                                 )
        $$
        Each sublist contains the [XX, YY, ZZ, Z] coefficients in this order.
        The last sublist should have the same shape, but only the Z component
        is used.
        If no coefficient list is provided, all are set to 1.
    '''

    # Three qubits because for 2 we get H_O = 0
    assert n_qubits >= 3

    if coeff == None:
        'Setting default values for the coefficients'
        coeff = [[1.0, 1.0, 1.0, 1.0] for i in range(n_qubits)]

    # Even terms of the Hamiltonian
    # (summing over individual pair-wise elements)
    H_E = sum((get_hamiltonian_n_site_terms(i, coeff[i], n_qubits)
               for i in range(0, n_qubits-1, 2)))

    # Odd terms of the Hamiltonian
    # (summing over individual pair-wise elements)
    H_O = sum((get_hamiltonian_n_site_terms(i, coeff[i], n_qubits)
               for i in range(1, n_qubits-1, 2)))

    # adding final Z term at the Nth site
    final_term = SparsePauliOp("I" * (n_qubits - 1) + "Z")
    final_term *= coeff[n_qubits-1][3]
    if (n_qubits % 2) == 0:
        H_E += final_term
    else:
        H_O += final_term

    # Returns the list of the two sets of terms
    return [H_E, H_O]


def get_time_evolution_operator(num_qubits, tau, trotter_steps, coeff=None):
    '''
        Given a number of qubits, generates the corresponding time-evolution for
        the Ising model with the same number of sites.

        Input:
            num_qubits (int): number of qubits, which should be equal to the
                number of spins in the chain
            evo_time (float): time parameter in time-evolution operator
            trotter_steps (int): number of time steps for the Suzuki-Trotter
                decomposition
            coeff (list of lists): parameters for each term in the Hamiltonian
                for each site ie ([[XX0, YY0, ZZ0, Z0], [XX1, YY1, ZZ1, Z1], ...])
        Returns:
            evo_op.definition: Trotterized time-evolution operator
    '''
    # Constructing the Hamiltonian here;
    # heisenberg_hamiltonian = [H_E, H_O]
    heisenberg_hamiltonian = get_heisenberg_hamiltonian(num_qubits,
                                                        coeff)

    # e^ (-i*H*evo_time), with Trotter decomposition
    # exp[(i * evo_time)*(IIIIXXIIII + IIIIYYIIII + IIIIZZIIII + IIIIZIIIII)]
    evo_op = PauliEvolutionGate(heisenberg_hamiltonian, tau,
                                synthesis=SuzukiTrotter(order=2,
                                    reps=trotter_steps))
    # The Trotter order=2 applies one set of the operators for
    # half a timestep, then the other set for a full timestep,
    # then the first step for another half a step note that reps
    # includes the number of repetitions of the Trotterized
    # operator higher number means more repetitions, and thus
    # allowing larger timestep
    return evo_op.definition


def find_string_pattern(pattern, string):
    match_list = []
    for m in re.finditer(pattern, string):
        match_list.append(m.start())
    return match_list


# efficient propagator for pauli hamiltonians
def sort_Pauli_by_symmetry(ham):
    '''
        Separates a qiskit PauliOp object terms into 1 and 2-qubit
        operators. Furthermore, 2-qubit operators are separated according
        to the parity of the index first non-identity operation.
    '''
    one_qubit_terms = []
    two_qubit_terms = []
    # separating the one-qubit from two-qubit terms
    for term in ham:
        matches = find_string_pattern('X|Y|Z', str(term.paulis[0]))
        pauli_string = term.paulis[0]
        coeff = np.real(term.coeffs[0])
        str_tag = pauli_string.to_label().replace('I', '')
        if len(matches) == 2:
            two_qubit_terms.append((pauli_string, coeff, matches, str_tag))
        elif len(matches) == 1:
            one_qubit_terms.append((pauli_string, coeff, matches, str_tag))

    # sorting the two-qubit terms according to index on which they act
    two_qubit_terms = sorted(two_qubit_terms, key=lambda x: x[2])
    # separating the even from the odd two-qubit terms
    even_two_qubit_terms = list(filter(lambda x: not x[2][0]%2, two_qubit_terms))
    odd_two_qubit_terms = list(filter(lambda x: x[2][0]%2, two_qubit_terms))

    even_two_qubit_terms = [list(v) for i, v in groupby(even_two_qubit_terms, lambda x: x[2][0])]
    odd_two_qubit_terms = [list(v) for i, v in groupby(odd_two_qubit_terms, lambda x: x[2][0])]

    return one_qubit_terms, even_two_qubit_terms, odd_two_qubit_terms


def generate_circ_pattern_1qubit(circ, term, delta_t):
    '''
        General 1-qubit gate for exponential of product identity and
        a single pauli gate.
        Only a single rotation operation is required, with the angle
        being related to the exponential argument:

        R_P(coeff) = exp(-i * coeff * P / 2)

        Where P is the Pauli gate and coeff encompasses the constant
        coefficient term
    '''
    coeff = 2 * term[1] * delta_t
    if term[3] == 'X':
        circ.rx(coeff, term[2])
    elif term[3] == 'Y':
        circ.ry(coeff, term[2])
    elif term[3] == 'Z':
        circ.rz(coeff, term[2])

    return circ


def generate_circ_pattern_2qubit(circ, term, delta_t):
    r'''
        General 2-qubit gate for exponential of Paulis. This is the
        optimal decomposition, based on a component of a U(4) operator.
        (see )

        The circuit structure is as follows:

        - ---- I ---- C - Rz(o) - X --- I --- C - Rz(pi/2) -
        - Rz(-pi/2) - X - Ry(p) - C - Ry(l) - X ---- I -----

        Where CX represent CNOT operations, R are rotation gates with angles,
        and I is the identity matrix. The angles are parameterized as follows:

        $ o = \theta = (\pi/2 - A) $
        $ p = \phi = (A - \pi/2) $
        $ l = \lambda = (\pi/2 - A) $

        Where A is the exponential argument.
    '''
    # wires to which to apply the operation
    wires = term[0][2]

    # angles to parameterize the circuit,
    # based on exponential argument
    if any('XX' in sublist for sublist in term):
        g_phi = ( 2 * (-1) * term[0][1] * delta_t - np.pi / 2)
    else:
        g_phi = - np.pi / 2
    if any('YY' in sublist for sublist in term):
        g_lambda = (np.pi/2 - 2 * (-1) * term[1][1] * delta_t)
    else:
        g_lambda = np.pi/2
    if any('ZZ' in sublist for sublist in term):
        g_theta = (np.pi/2 - 2 * (-1) * term[2][1] * delta_t)
    else:
        g_theta = np.pi/2

    # circuit
    circ.rz(-np.pi/2, wires[1])
    circ.cx(wires[1], wires[0])
    circ.rz(g_theta, wires[0])
    circ.ry(g_phi, wires[1])
    circ.cx(wires[0], wires[1])
    circ.ry(g_lambda, wires[1])
    circ.cx(wires[1], wires[0])
    circ.rz(np.pi/2, wires[0])
    return circ


def get_manual_Trotter(num_q, pauli_ops, timestep, n_trotter=1,
                       trotter_type='basic', reverse_bits=True):
    # sorts the Pauli strings according to qubit number they affect and symmetry
    one_q, even_two_q, odd_two_q = sort_Pauli_by_symmetry(pauli_ops)
    # scales the timestep according to the number of trotter steps
    timestep_even_two_q = timestep / n_trotter
    timestep_odd_two_q = timestep / n_trotter
    timestep_one_q = timestep / n_trotter
    # symmetric places 1/2 of one_q and odd_two_q before and after even_two_q
    if trotter_type == 'symmetric':
        timestep_odd_two_q /= 2
        timestep_one_q /= 2
    # constructs circuits for each segment of the operators
    qc_odd_two_q, qc_even_two_q, qc_one_q = QuantumCircuit(num_q), QuantumCircuit(num_q), QuantumCircuit(num_q)
    for i in even_two_q:
        qc_even_two_q = generate_circ_pattern_2qubit(qc_even_two_q, i, timestep_even_two_q)
    for i in odd_two_q:
        qc_odd_two_q = generate_circ_pattern_2qubit(qc_odd_two_q, i, timestep_odd_two_q)
    for i in one_q:
        qc_one_q = generate_circ_pattern_1qubit(qc_one_q, i, timestep_one_q)
    # assembles the circuit for Trotter decomposition of exponential
    qr = QuantumRegister(num_q)
    qc = QuantumCircuit(qr)
    if trotter_type == 'basic':
        qc = qc.compose(qc_even_two_q)
        qc = qc.compose(qc_odd_two_q)
        qc = qc.compose(qc_one_q)
    elif trotter_type == 'symmetric':
        qc = qc.compose(qc_one_q)
        qc = qc.compose(qc_odd_two_q)
        qc = qc.compose(qc_even_two_q)
        qc = qc.compose(qc_odd_two_q)
        qc = qc.compose(qc_one_q)
    # repeats the single_trotter circuit several times to match n_trotter
    for i in range(n_trotter-1):
        qc = qc.compose(qc)
    if reverse_bits:
        return qc.reverse_bits()
    else:
        return qc


if __name__ == '__main__':
    num_shots = 100
    num_q = 3
    evolution_timestep = 0.1
    n_trotter_steps = 1
    # XX YY ZZ, Z
    ham_coeffs = ([[0.75/2, 0.75/2, 0.0, 0.65]]
                    + [[0.5, 0.5, 0.0, 1.0]
                    for i in range(num_q-1)])
    time_evo_op = get_time_evolution_operator(
        num_qubits=num_q, tau=evolution_timestep,
        trotter_steps=n_trotter_steps, coeff=ham_coeffs)
    print(time_evo_op)

    spin_chain_hamiltonian = get_heisenberg_hamiltonian(num_q,
                                                        ham_coeffs)

    spin_chain_hamiltonian = sum(spin_chain_hamiltonian)
    print(get_manual_Trotter(num_q, spin_chain_hamiltonian,
                             0.1).draw())
    print(get_manual_Trotter(num_q, spin_chain_hamiltonian, 0.1,
                             n_trotter=2).draw())
    print(get_manual_Trotter(num_q, spin_chain_hamiltonian, 0.1,
                             trotter_type='symmetric').draw())
    print(get_manual_Trotter(num_q, spin_chain_hamiltonian, 0.1,
                             n_trotter=2,
                             trotter_type='symmetric').draw())



import numpy as np
import matplotlib.pyplot as plt
from .utils import execute
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
from qiskit_aer import Aer
from .spin_propagators import get_time_evolution_operator

class SpinDynamicsS:
    """
    A class to simulate the dynamics of a quantum system using a statevector approach.

    Attributes:
        num_qubits (int): Number of qubits in the quantum system.
        evolution_timestep (float): Time step for the evolution.
        trotter_steps (int): Number of Trotter steps for the simulation.
        hamiltonian_coefficients (list): Hamiltonian coefficients for the system.
        initial_state (str or list): Initial state of the system, represented as a binary string or list.
        time_evo_op (QuantumCircuit): Time evolution operator for the system.
        psin0 (ndarray): Initial state vector.
        psin_list (list): List of state vectors during the simulation.
        correlation_list (list): List of correlations calculated during the simulation.
        dpi (int): DPI setting for plot output.
    """

    def __init__(
        self,
        num_qubits,
        evolution_timestep,
        trotter_steps,
        hamiltonian_coefficients,
    ):
        """
        Initialize the QuantumDynamicsClassicalSimulation class.

        Args:
            num_qubits (int): Number of qubits in the quantum system.
            evolution_timestep (float): Time step for the evolution.
            trotter_steps (int): Number of Trotter steps for the simulation.
            hamiltonian_coefficients (list): Hamiltonian coefficients for the system.
            initial_state (str or list): Initial state of the system, represented as a binary string or list.
        """
        self.num_qubits = num_qubits
        self.evolution_timestep = evolution_timestep
        self.trotter_steps = trotter_steps
        self.hamiltonian_coefficients = hamiltonian_coefficients
        self.time_evo_op = get_time_evolution_operator(
            num_qubits=self.num_qubits,
            tau=self.evolution_timestep,
            trotter_steps=self.trotter_steps,
            coeff=self.hamiltonian_coefficients,
        )
        self.initial_state = ''
        self.psin_list = []
        self.correlation_list = []

        # Plot settings
        self.dpi = 300
        plt.rcParams["axes.linewidth"] = 1.5
        plt.rcParams["lines.markersize"] = 11
        plt.rcParams["figure.figsize"] = (6.4, 3.6)

    def prepare_initial_state(self, state_string):
        """
        Prepare the initial state vector from the binary string or list.

        Returns:
            ndarray: Flattened initial state vector.
        """
        zero_state = np.array([[1], [0]])
        one_state = np.array([[0], [1]])

        # Map binary string or list to quantum states
        state_vectors = [
            zero_state if str(bit) == "0" else one_state for bit in state_string
        ]

        # Perform Kronecker product to construct the full initial state
        psin = state_vectors[0]
        for state in state_vectors[1:]:
            psin = np.kron(psin, state)
        return psin.flatten()

    def qsolve_statevector(self, psin):
        """
        Perform statevector propagation for the quantum circuit.

        Args:
            psin (ndarray): Input statevector.

        Returns:
            ndarray: Final statevector after execution.
        """
        d = int(np.log2(np.size(psin)))
        qre = QuantumRegister(d)
        circ = QuantumCircuit(qre)
        circ.initialize(psin, qre)
        circ.barrier()
        circ.append(self.time_evo_op, qre)
        circ.barrier()

        device = Aer.get_backend("statevector_simulator")
        result = execute(circ, backend=device).result()
        return result.get_statevector()

    def run_dynamics(self, nsteps, state_string):
        """
        Simulate the dynamics of the quantum system over a number of steps.

        Args:
            nsteps (int): Number of time steps for the simulation.
        """
        self.psin0 = self.prepare_initial_state(state_string)
        self.psin_list = [self.psin0]
        for k in range(nsteps):
            print(f"Running dynamics step {k}")
            if k > 0:
                psin = self.qsolve_statevector(self.psin_list[-1])
                self.psin_list.pop()
                self.psin_list.append(psin)

            self.correlation_list.append(np.vdot(self.psin_list[-1], self.psin0))

    def save_results(self, filename_prefix):
        """
        Save the simulation results to files.

        Args:
            filename_prefix (str): Prefix for the output filenames.
        """
        time = np.arange(
            0,
            self.evolution_timestep * len(self.correlation_list),
            self.evolution_timestep,
        )
        np.save(f"{filename_prefix}_time", time)
        sa_observable = np.abs(self.correlation_list)
        np.save(f"{filename_prefix}_SA_obs", sa_observable)

    def plot_results(self, filename_prefix):
        """
        Plot the simulation results and save the plots as files.

        Args:
            filename_prefix (str): Prefix for the output filenames.
        """
        time = np.arange(
            0,
            self.evolution_timestep * len(self.correlation_list),
            self.evolution_timestep,
        )
        sa_observable = np.abs(self.correlation_list)

        plt.plot(time, sa_observable, "-o")
        plt.xlabel("Time")
        plt.ylabel(r"$\left|\langle \psi | \psi (t)  \rangle \right|$")
        plt.xlim((min(time), max(time)))
        plt.yscale("log")
        #plt.legend()
        plt.tight_layout()
        plt.savefig(f"{filename_prefix}.pdf", format="pdf", dpi=self.dpi)
        plt.savefig(f"{filename_prefix}.png", format="png", dpi=self.dpi)
        plt.show()


class SpinDynamicsH:
    """
    A class to simulate quantum dynamics using Hadamard tests and time evolution operators.

    This class performs quantum circuit construction, execution, and data processing to calculate
    survival amplitudes and probabilities for a spin chain system.
    """
    def __init__(self, num_qubits, evolution_timestep, trotter_steps, hamiltonian_coefficients):
        """
        Initializes the QuantumSimulation class with system parameters and sets up the simulator.

        Parameters:
            num_qubits (int): Number of qubits in the system.
            hamiltonian_coefficients (list): Coefficients for the Hamiltonian terms.
            evolution_timestep (float): Timestep for the time evolution operator.
        """
        self.num_qubits = num_qubits
        self.hamiltonian_coefficients = hamiltonian_coefficients
        self.evolution_timestep = evolution_timestep
        self.trotter_steps = trotter_steps,
        self.simulator = Aer.get_backend('qasm_simulator')
        self.real_amp_list = []
        self.imag_amp_list = []

        self.time_evo_op = get_time_evolution_operator(
            num_qubits=num_qubits,
            tau=evolution_timestep,
            trotter_steps=trotter_steps,
            coeff=hamiltonian_coefficients
        )
        self.controlled_time_evo_op = self.time_evo_op.control()

    def get_hadamard_test(self, initial_state, control_repeats, imag_expectation=False):
        """
        Constructs the Hadamard test circuit to evaluate the real or imaginary components of the operator.

        Parameters:
            initial_state (QuantumCircuit): Circuit for initializing the quantum state.
            control_repeats (int): Number of times the controlled operation is applied.
            imag_expectation (bool): Whether to evaluate the imaginary component (default: False).

        Returns:
            QuantumCircuit: The constructed Hadamard test circuit.
        """
        qr = QuantumRegister(self.num_qubits + 1)
        cr = ClassicalRegister(1)
        qc = QuantumCircuit(qr, cr)
        qc.append(initial_state, qr[1:])
        qc.barrier()

        qc.h(0)
        if imag_expectation:
            qc.p(-np.pi / 2, 0)

        for _ in range(control_repeats):
            qc.append(self.controlled_time_evo_op, qr[:])

        qc.h(0)
        qc.barrier()
        qc.measure(0, 0)
        return qc

    def execute_circuit(self, qc, num_shots=100):
        """
        Executes a quantum circuit using the Qiskit simulator and retrieves measurement counts.

        Parameters:
            qc (QuantumCircuit): The quantum circuit to execute.
            num_shots (int): Number of shots for circuit execution (default: 100).

        Returns:
            dict: Measurement counts.
        """
        job = execute(qc, self.simulator, shots=num_shots)
        return job.result().get_counts()

    @staticmethod
    def calculate_spin_correlation(counts):
        """
        Calculates the spin correlation based on measurement counts.

        Parameters:
            counts (dict): Measurement counts.

        Returns:
            float: Average spin correlation.
        """
        qubit_to_spin_map = {'0': 1, '1': -1}
        total_counts = sum(counts.values())
        values = [qubit_to_spin_map[k] * v for k, v in counts.items()]
        return sum(values) / total_counts

    @staticmethod
    def initialize_state(num_qubits, state_string):
        """
        Creates a circuit to initialize the quantum state.

        Parameters:
            num_qubits (int): Number of qubits.
            state_string (str): Binary string representing the initial state.

        Returns:
            QuantumCircuit: Circuit for initializing the state.
        """
        state_string = ''.join('1' if char == '0' else '0' for char in state_string)
        qr = QuantumRegister(num_qubits)
        qc = QuantumCircuit(qr)
        qc.initialize(state_string, qr[:])
        return qc

    def run_simulation(self, state_string, total_time, num_shots=100):
        """
        Runs the Hadamard test simulation for the given initial state.

        Parameters:
            state_string (str): Binary string representing the initial state.
            total_time (float): Total simulation time.
            num_shots (int): Number of shots for circuit execution (default: 100).
        """
        init_circuit = self.initialize_state(self.num_qubits, state_string)
        self.time_range = np.arange(0, total_time + self.evolution_timestep,
                                    self.evolution_timestep)

        for idx, _ in enumerate(self.time_range):
            print(f'Running dynamics step {idx}')

            # Real component
            qc_real = self.get_hadamard_test(init_circuit, idx, imag_expectation=False)
            real_counts = self.execute_circuit(qc_real, num_shots)
            real_amp = self.calculate_spin_correlation(real_counts)
            self.real_amp_list.append(real_amp)

            # Imaginary component
            qc_imag = self.get_hadamard_test(init_circuit, idx, imag_expectation=True)
            imag_counts = self.execute_circuit(qc_imag, num_shots)
            imag_amp = self.calculate_spin_correlation(imag_counts)
            self.imag_amp_list.append(imag_amp)

            print(f'Finished step {idx}: Re = {real_amp:.3f}, Im = {imag_amp:.3f}')

    def save_results(self, prefix):
        """
        Saves the real and imaginary amplitudes, survival amplitude, and survival probability to files.

        Parameters:
            prefix (str): Prefix for the output file names.
        """
        real_amp_array = np.array(self.real_amp_list)
        imag_amp_array = np.array(self.imag_amp_list)

        np.savetxt(f'{prefix}_real_amp.csv', real_amp_array, fmt='%.18e', delimiter=';')
        np.savetxt(f'{prefix}_imag_amp.csv', imag_amp_array, fmt='%.18e', delimiter=';')
        np.savetxt(f'{prefix}_abs_correlation.csv', np.abs(real_amp_array + 1j * imag_amp_array), fmt='%.18e', delimiter=';')
        np.savetxt(f'{prefix}_sqrt_sum_squares.csv', np.sqrt(real_amp_array**2 + imag_amp_array**2), fmt='%.18e', delimiter=';')

    def plot_results(self, prefix):
        """
        Plots the survival amplitude and compares it with a reference.

        Parameters:
            prefix (str): Prefix for loading precomputed reference data.
        """
        abs_corr = np.abs(np.array(self.real_amp_list) + 1j * np.array(self.imag_amp_list))

        plt.plot(self.time_range, abs_corr, '.', label='Hadamard Test')
        ref_sa = np.load(f'../../../data/{self.num_qubits}_spin_chain_SA_obs.npy')
        ref_time = np.load(f'../../../data/{self.num_qubits}_spin_chain_time.npy')
        plt.plot(ref_time, ref_sa, '-', label='Statevector')

        plt.xlabel('Time')
        plt.ylabel(r"$\left|\langle \psi | \psi (t)  \rangle \right|$")
        plt.tight_layout()
        #plt.legend()
        plt.show()


if __name__ == "__main__":
    classical = True
    num_q = 3
    evolution_timestep = 0.1
    n_trotter_steps = 1
    hamiltonian_coefficients = [[0.75 / 2, 0.75 / 2, 0.0, 0.65]] + [
        [0.5, 0.5, 0.0, 1.0] for _ in range(num_q - 1)
    ]
    initial_state = "011"  # Specify the initial state as a binary string

    if classical:
        csimulation = SpinDynamicsS(
            num_q,
            evolution_timestep,
            n_trotter_steps,
            hamiltonian_coefficients,
        )
        csimulation.run_dynamics(nsteps=250, state_string=initial_state)
        csimulation.save_results(f"{num_q}_spin_chain")
        csimulation.plot_results(f"{num_q}_spin_chain_statevector")

    else:
        qsimulation = SpinDynamicsH(
            num_q,
            evolution_timestep,
            n_trotter_steps,
            hamiltonian_coefficients,
        )
        qsimulation.run_simulation(state_string=initial_state, total_time=25, num_shots=100)
        qsimulation.save_results('hadamard_test')
        qsimulation.plot_results('hadamard_test')



import numpy as np
from scipy.linalg import expm, kron
 

def vectorize_comm(A):
    # Create an identity matrix with the same dimension as A
    iden = np.eye(A.shape[0])
    # Compute the vectorized commutator [A, .] as the Kronecker product 
    return kron(iden, A) - kron(A.T, iden)

class VectorizedEffectiveHamiltonian_class:
    def __init__(self, He, Ha):
        self.He = He
        self.Ha = Ha

def VectorizedEffectiveHamiltonian(H, gamma, lind):
    # Create an identity matrix with the same dimension as H
    iden = np.eye(H.shape[0])
    # Get the dimension of H
    d = H.shape[0]
    # Compute the vectorized commutator for the Hamiltonian H
    vec_H = vectorize_comm(H)
    # Initialize the result matrix with zeros (complex type)
    res = np.zeros((d**2, d**2), dtype=np.complex128)
    
    # Compute the conjugate of the Lindblad operator
    L_conj = lind.conj()
    L_dagger_L = L_conj.T @ lind
    # Compute the Lindblad contribution to the effective Hamiltonian
    res -= gamma * (kron(L_conj, lind) - (kron(iden, L_dagger_L) + kron(L_dagger_L.T, iden)) / 2)

    # Return an instance of the VectorizedEffectiveHamiltonian_class with vec_H and res
    return VectorizedEffectiveHamiltonian_class(vec_H, res)

class EffectiveHamiltonian_class:
    def __init__(self, He, Ha, Llist, LdL):
        self.He = He  # Hermitian part
        self.Ha = Ha  # Anti-Hermitian part
        self.Llist = Llist  # List of Lindblad operators
        self.LdL = LdL  # List of L†L

def EffectiveHamiltonian( mats, Llist):
    """
    Create an EffectiveHamiltonian object based on provided parameters.
    :param mats: List of matrices (Hamiltonian terms).
    :param Llist: List of lists of Lindblad operators.
    :return: An instance of EffectiveHamiltonian_class.
    """
    # Directly use the input matrices as the Hermitian part (assuming they are Hermitian)
    He = sum(mats)  # Sum of Hamiltonian terms as Hermitian part
    Ha = 0.0  # Initialize anti-Hermitian part
    LdL = []  # Initialize the list for Lindblad operator products

    for  LL in Llist:
        for L in LL:
            L_dagger_L = (L.conj().T @ L) 
            
            LdL.append(L_dagger_L)  # Append to LdL list
            Ha += L_dagger_L  # Sum for the anti-Hermitian part

    # Return the Effective Hamiltonian object
    return EffectiveHamiltonian_class(He, 0.5 * Ha, [L for LL in Llist for L in LL], LdL)

# Helper methods that align with the Julia logic
def herm(H, t):
    return H.He(t)  # Assuming He is callable with t

def antiherm(H):
    return H.Ha

def get_LdL(H):
    return H.LdL

def get_L(H):
    return H.Llist



# Utilities
# Plotting utilities, conversion factors, etc.
import numpy as np
import qiskit_aer
from qiskit.compiler import transpile
from qiskit_ibm_runtime import Sampler
from qiskit.quantum_info import SparsePauliOp
import itertools

# Conversion Factors:
def convert_au_to_eV(input_val):
    au2ev = 27.21138602
    return(au2ev * input_val)


def convert_eV_to_au(input_val):
    au2ev = 27.21138602
    ev2au = 1/au2ev
    return(ev2au * input_val)


def convert_bohr_to_au(input_val):
    bohr2au = 0.52917721092
    return(bohr2au * input_val)


def convert_au_to_bohr(input_val):
    bohr2au = 0.52917721092
    au2bohr = 1/bohr2au
    return(input_val * au2bohr)


def convert_fs_to_au(input_val):
    fs2au = 41.3414
    return(fs2au * input_val)


def convert_au_to_fs(input_val):
    fs2au = 41.3414
    au2fs = 1/fs2au
    return(au2fs * input_val)


def get_proton_mass():
    proton_mass = 1836.15267343 # proton-electron mass ratio
    return(proton_mass)


def execute(QCircuit, backend=None, shots=None, real_backend=False):
    '''
        Function to replace the now-deprecated Qiskit
        `QuantumCircuit.execute()` method.

        Input:
          - `QCircuit`: qiskit.QuantumCircuit object
          - `Backend`: qiskit.Backend instance
          - `shots`: int specifying the number of shots
          - `real_backend`: bool specifying whether the provided backend is
            a real device (True) or not (False)
    '''
    if shots:
        n_shots = shots
    else:
        n_shots = 1024 # Use the qiskit default if not specified

    if real_backend:
        QCircuit.measure_all()
        qc = transpile(QCircuit, backend=backend)
        sampler = Sampler(backend)
        job = sampler.run([qc], shots=n_shots)
    else:
        # Transpile circuit with statevector backend
        tmp_circuit = transpile(QCircuit, backend)
        # Run the transpiled circuit
        job = backend.run(tmp_circuit, n_shots=shots)
    return(job)


# Calculation of Expectation Value:
def calculate_expectation_values(dynamics_results, observable_grid, do_FFT=False, dx=None):
    '''
    Function to calculate the time-dependent expectation value of an observable O defined on a grid.
    Inputs:

        - `dynamics_results`: np.ndarray of wavefunctions/propagated states with shape: (n_steps, nx)
        - `observable_grid`: np.array of observable
    '''
    if dx:
        d_observable = dx
    else:
        d_observable = observable_grid[1] - observable_grid[0]
    if do_FFT:
        psi_list = np.fft.fft(dynamics_results, axis=1, norm='ortho')
    else:
        psi_list = dynamics_results
    # Compute the expectation value.
    expectation  = np.real(np.sum(psi_list.conj()*observable_grid*psi_list*d_observable, axis=1))

    return(expectation)

# Pauli Decomposition Utilities

def vec_query(arr, my_dict):
    '''
    This function vectorizes dictionary querying, allowing us to query `my_dict` with a np.array `arr` of keys.
    '''

    return np.vectorize(my_dict.__getitem__, otypes=[tuple])(arr)


def nested_kronecker_product(a):
    '''
    Handles Kronecker Products for list (i.e.,  a = [Z, Z, Z] will evaluate Z ⊗ Z ⊗ Z)
    '''
    if len(a) == 2:
            return np.kron(a[0],a[1])
    else:
        return np.kron(a[0], nested_kronecker_product(a[1:]))


def Hilbert_Schmidt(mat1, mat2):
    'Return the Hilbert-Schmidt Inner Product of two matrices.'
    return np.trace(mat1.conj().T * mat2)


def decompose(Ham_arr, tol=12, subset = None):
    '''
    Function that takes an input matrix `H` and decomposes into a sum of tensor products of Pauli Matrices.
        - The expectation is that `H` will have a shape of 2^n, where n is the number of qubits needed to represent
        the matrix in this form. n determines the number of terms in the tensor product composing the Pauli strings.
        - Ex: If H has shape of (16, 16), n = log2(H) = 4 qubits.
         The Pauli strings will then take the form of ['IIII', 'ZIII', 'ZZII', etc.]
    Has the option to toggle verbose output, which prints the terms of the pauli sum.
    '''

    X = np.asarray([[0,1],[1,0]])
    Y = np.asarray([[0,complex(0,-1)],[complex(0,1),0]])
    Z = np.asarray([[1,0],[0,-1]])
    I = Z@Z
    # Define a dictionary with the four Pauli matrices:
    global_pms = {'I': I,'X': X,'Y': Y,'Z': Z}

    # The next few lines allow for the function use a reduced basis.
    pm_keys = []
    pm_vals = []
    if subset:
        for tmp_key in subset:
            pm_keys.append(tmp_key)
            pm_vals.append(global_pms[tmp_key])
        pms = dict(zip(pm_keys, pm_vals))
    else:
        pms = global_pms

    pauli_keys = list(pms.keys()) # Keys of the dictionary

    nqb = int(np.log2(Ham_arr.shape[0])) # Determine the # of qubits needed

    # Make all possible tensor products of Pauli matrices sigma
    sigma_combinations = list(itertools.product(pauli_keys, repeat=nqb))

    output_string = '' # Initialize an empty string to which we can add our terms
    for ii in range(len(sigma_combinations)):
        pauli_str = ''.join(sigma_combinations[ii])

        # Convert the Pauli string into a list of matrices
        tmp_mat_list = vec_query(np.array(sigma_combinations[ii]), pms)

        # Evaluate the Kronecker product of the matrix array
        tmp_p_matrix = nested_kronecker_product(tmp_mat_list)

        # Compute the coefficient for each Pauli string
        a_coeff = (1/(2**nqb)) * Hilbert_Schmidt(tmp_p_matrix, Ham_arr)

        # If the coefficient is non-zero, we want to use it!
        if abs(a_coeff) > 10**(-tol):
            output_string += str(np.round(a_coeff.real, tol))+'*'+pauli_str
            output_string += '+' # Add a plus sign for the next term!

    return output_string[:-1] # To ignore that extra plus sign


def build_pauli_dict(decomposed_operator):
    '''
    Build a Pauli dictionary from the output of the `decompose` function above.
    This is done by converting the unique Pauli Strings ['IIII', 'IIIZ', etc.] to keys,
    where the values of the dictionary are the numerical coefficients.
    '''
    single_terms = decomposed_operator.split('+')
    pauli_dict_out = {}
    for term in single_terms:
        coeff, pauli_str = term.split('*')
        pauli_dict_out[pauli_str] = float(coeff)
    return pauli_dict_out


def pauli_strings_2_pauli_sum(operator):
    '''
    Function to convert the output of the `decompose` function above into a Qiskit-recognized PauliSumOp.
    The string representing the Hamiltonian as a sum of product of Paulis is converted into a dictionary where
    the keys are the Pauli Strings (ex: 'IIII' or 'IXYZ') and the values are the coefficients.
    '''
    tmp_pauli_dict = build_pauli_dict(operator)
    # Convert the dict to a list of tuples of the form [('Pauli String', float_coeff), ...]
    tmp_pauli_sum = SparsePauliOp(data=list(tmp_pauli_dict.keys()), coeffs=list(tmp_pauli_dict.values()))
    return tmp_pauli_sum



import numpy as np
import scipy.linalg as LA
import scipy.fft as sfft
from qutip import mesolve, Qobj
from typing import List, Tuple, Callable, Optional, Any

from . import params as pa
from . import trans_basis as tb


class DynamicsOS:
    """Class for open-system dynamics (Lindblad equation).

    This class provides methods to simulate open-system dynamics described by the Lindblad equation.

    Attributes:
        Nsys (int): System Hilbert Space Dimension.
        Hsys (np.ndarray): Hamiltonian of the system (shape (N, N)).
        rho0 (np.ndarray): Initial density matrix (shape (N, N)).
        c_ops (List[np.ndarray]): List of collapse operators (each of shape (N, N)).
    """

    def __init__(
        self,
        Nsys: int,
        Hsys: np.ndarray,
        rho0: np.ndarray,
        c_ops: Optional[List[np.ndarray]] = None
    ) -> None:
        """
        Initialize the DynamicsOS instance.

        Args:
            Nsys (int): System Hilbert Space Dimension.
            Hsys (np.ndarray): Hamiltonian of the system.
            rho0 (np.ndarray): Initial density matrix.
            c_ops (Optional[List[np.ndarray]]): List of collapse operators. Defaults to an empty list.
        """
        if c_ops is None:
            c_ops = []
        self.Nsys: int = Nsys
        self.Hsys: np.ndarray = Hsys
        self.rho0: np.ndarray = rho0
        self.c_ops: List[np.ndarray] = c_ops

    def Gt_matrix_expo(self, time_arr: List[float], Is_show_step: bool = False) -> List[np.ndarray]:
        """
        Compute the propagator of the Lindblad equation using the matrix exponential.

        The propagator is computed by exponentiating the Liouvillian operator defined by the
        system Hamiltonian and collapse operators.

        Args:
            time_arr (List[float]): Array of time values for the simulation.
            Is_show_step (bool, optional): If True, prints the current simulation step. Defaults to False.

        Returns:
            List[np.ndarray]: List of propagators corresponding to each time in `time_arr`.
        """
        ident_h: np.ndarray = np.eye(self.Nsys, dtype=np.complex128)

        # Build the A matrix for time-derivation of the vectorized density matrix.
        Amat: np.ndarray = -1j * (np.kron(self.Hsys, ident_h) - np.kron(ident_h, self.Hsys.T))
        for i in range(len(self.c_ops)):
            Amat += 0.5 * (
                2.0 * np.kron(self.c_ops[i], self.c_ops[i].conj())
                - np.kron(ident_h, (self.c_ops[i].T @ self.c_ops[i].conj()))
                - np.kron((self.c_ops[i].T.conj() @ self.c_ops[i]), ident_h)
            )

        G_prop: List[np.ndarray] = []
        for i, t in enumerate(time_arr):
            if Is_show_step:
                print("step", i, "time", t)
            Gt: np.ndarray = LA.expm(Amat * t)
            G_prop.append(Gt)
        return G_prop

    def propagate_matrix_exp(
        self,
        time_arr: List[float],
        observable: np.ndarray,
        Is_store_state: bool = False,
        Is_show_step: bool = False,
        Is_Gt: bool = False,
    ) -> Any:
        """
        Solve the Lindblad equation using matrix exponential.

        This method computes the propagator, evolves the initial density matrix, and calculates
        the expectation value of the observable over time. Optionally, it stores the evolved density matrices.

        Args:
            time_arr (List[float]): Time array for dynamic simulation.
            observable (np.ndarray): Observable matrix for which the expectation value is computed.
            Is_store_state (bool, optional): If True, stores the density matrices at each time step.
                                             Defaults to False.
            Is_show_step (bool, optional): If True, prints the current simulation step. Defaults to False.
            Is_Gt (bool, optional): If True, includes the propagators in the result. Defaults to False.

        Returns:
            Result: An object with the following attributes:
                - expect (List[float]): List of expectation values over time.
                - density_matrix (List[np.ndarray], optional): List of density matrices (if `Is_store_state` is True).
                - Gprop (List[np.ndarray], optional): List of propagators (if `Is_Gt` is True).
        """

        class Result:
            """Class for storing propagation results."""
            def __init__(self, store_state: bool, include_Gt: bool) -> None:
                self.expect: List[float] = []
                if store_state:
                    self.density_matrix: List[np.ndarray] = []
                if include_Gt:
                    self.Gprop: Optional[List[np.ndarray]] = None

        result = Result(Is_store_state, Is_Gt)

        # Compute the propagator of the Lindblad equation.
        G_prop: List[np.ndarray] = self.Gt_matrix_expo(time_arr, Is_show_step)
        if Is_Gt:
            result.Gprop = G_prop

        # Initialize the vectorized density matrix.
        vec_rho0: np.ndarray = self.rho0.reshape(self.Nsys**2)

        for i, _ in enumerate(time_arr):
            vec_rhot: np.ndarray = G_prop[i] @ vec_rho0
            # Reshape back to density matrix form.
            rhot: np.ndarray = vec_rhot.reshape(self.Nsys, self.Nsys)

            if Is_store_state:
                result.density_matrix.append(rhot)
            result.expect.append(np.trace(rhot @ observable).real)

        return result

    def propagate_qt(self, time_arr: List[float], observable: Any, **kwargs: Any) -> List[float]:
        """
        Propagate the system using QuTiP's `mesolve` function.

        This method solves the Lindblad master equation using QuTiP's `mesolve` to compute the expectation
        values of the observable over time.

        Args:
            time_arr (List[float]): Time array for dynamic simulation.
            observable (Any): Observable operator(s) for which the expectation value is computed.
                              Can be a single operator or a list of operators.
            **kwargs: Additional keyword arguments to pass to `mesolve`.

        Returns:
            List[float]: List of expectation values of the observable over time.
        """
        c_ops: List[Qobj] = [Qobj(c_op) for c_op in self.c_ops]

        if isinstance(observable, list):
            obs = [Qobj(op) for op in observable]
        else:
            obs = Qobj(observable)

        result = mesolve(Qobj(self.Hsys), Qobj(self.rho0), time_arr, c_ops, obs, **kwargs)
        return result.expect


class DVR_grid:
    """Class for Discrete Variable Representation (DVR) grid methods.

    This class handles grid-based representations for systems where the potential is expressed on grid points.

    Attributes:
        Ngrid (int): Number of grid points.
        xmin (float): Minimum value of the grid.
        xmax (float): Maximum value of the grid.
        mass (float): Mass of the particle.
        xgrid (np.ndarray): Array of grid points in position space.
        dx (float): Spacing between grid points.
        dk (float): Spacing in momentum space.
        kgrid (np.ndarray): Array of grid points in momentum space.
        ak2 (np.ndarray): Kinetic energy array in momentum space.
        hamk (np.ndarray): Kinetic Hamiltonian matrix in position space.
        potential (Optional[np.ndarray]): Potential energy array on the grid.
    """

    def __init__(self, xmin: float, xmax: float, Ngrid: int, mass: float) -> None:
        """
        Initialize the DVR_grid instance.

        Args:
            xmin (float): Minimum x-value.
            xmax (float): Maximum x-value.
            Ngrid (int): Number of grid points.
            mass (float): Mass of the particle.
        """
        self.Ngrid: int = Ngrid
        self.xmin: float = xmin
        self.xmax: float = xmax
        self.mass: float = mass

        # Set up the position grid.
        self.xgrid: np.ndarray = np.array([])
        self._set_xgrid()
        self.dx: float = self.xgrid[1] - self.xgrid[0]

        # Set up the momentum grid.
        self.dk: float = 2.0 * np.pi / (self.Ngrid * self.dx)
        self.kgrid: np.ndarray = np.array([])
        self.ak2: np.ndarray = np.array([])  # Kinetic energy array.
        self.hamk: np.ndarray = np.array([])  # Kinetic Hamiltonian matrix.
        self._set_kinet_ham()

        # Potential energy array (to be set later).
        self.potential: Optional[np.ndarray] = None

    def _set_xgrid(self) -> None:
        """
        Set up the position space grid.

        Initializes the `xgrid` attribute using a linear space between `xmin` and `xmax`.
        """
        self.xgrid = np.linspace(self.xmin, self.xmax, self.Ngrid)

    def set_potential(self, func_pot: Callable[[float], float]) -> None:
        """
        Set up the potential energy array on the grid.

        Args:
            func_pot (Callable[[float], float]): Function that returns the potential value at a given x.
        """
        self.potential = np.zeros_like(self.xgrid)
        for i in range(self.Ngrid):
            self.potential[i] = func_pot(self.xgrid[i])

    def _set_kinet_ham(self) -> None:
        """
        Set up the kinetic Hamiltonian matrix in position space.

        This method computes the momentum grid and the corresponding kinetic energy values,
        and constructs the kinetic Hamiltonian matrix in position space using a Fourier transform.
        """
        self.kgrid = np.zeros(self.Ngrid, dtype=np.float64)
        self.ak2 = np.zeros(self.Ngrid, dtype=np.float64)

        coef_k: float = pa.hbar**2 / (2.0 * self.mass)

        for i in range(self.Ngrid):
            if i < self.Ngrid // 2:
                self.kgrid[i] = i * self.dk
            else:
                self.kgrid[i] = -(self.Ngrid - i) * self.dk
            self.ak2[i] = coef_k * self.kgrid[i]**2

        akx0: np.ndarray = sfft.ifft(self.ak2)
        self.hamk = np.zeros((self.Ngrid, self.Ngrid), dtype=np.complex128)

        for i in range(self.Ngrid):
            for j in range(self.Ngrid):
                if i < j:
                    self.hamk[i, j] = akx0[i - j].conj()
                else:
                    self.hamk[i, j] = akx0[i - j]

    def get_eig_state(self, Nstate: int) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get the eigenstates for the potential in x-space.

        Args:
            Nstate (int): Number of eigenstates to output.

        Returns:
            Tuple[np.ndarray, np.ndarray]: A tuple containing:
                - Eigenvalues (np.ndarray) for the first `Nstate` states.
                - Eigenvectors (np.ndarray) for the first `Nstate` states, normalized by sqrt(dx).
        """
        Mata: np.ndarray = self.hamk.copy()
        for i in range(self.Ngrid):
            Mata[i, i] += self.potential[i]

        val, arr = LA.eigh(Mata)
        return val[:Nstate], arr[:, :Nstate] / np.sqrt(self.dx)

    def x2k_wave(self, psi: np.ndarray) -> np.ndarray:
        """
        Transform the wavefunction from position space to momentum space.

        Args:
            psi (np.ndarray): Wavefunction in position space.

        Returns:
            np.ndarray: Wavefunction in momentum space.
        """
        return tb.x2k_wave(self.dx, psi)

    def k2x_wave(self, psik: np.ndarray) -> np.ndarray:
        """
        Transform the wavefunction from momentum space to position space.

        Args:
            psik (np.ndarray): Wavefunction in momentum space.

        Returns:
            np.ndarray: Wavefunction in position space.
        """
        return tb.k2x_wave(self.dx, psik)



"""
Class for GQME calculations 
"""

import numpy as np
import scipy.linalg as LA
from . import params as pa
from . import tt_tfd as tfd
from typing import Tuple, Optional

import time
import sys


class DynamicsGQME:
    """
    Class for Generalized Quantum Master Equation (GQME) calculation.
    
    This class provides methods for calculating the memory kernel, 
    performing TT-TFD calculations, and solving the GQME for open quantum systems.
    
    Attributes:
        Nsys (int): System Hilbert space dimension.
        Hsys (np.ndarray): System Hamiltonian of shape (N, N).
        rho0 (np.ndarray): Initial density matrix of shape (N, N).
        vec_rho0 (np.ndarray): Vectorized initial density matrix of shape (N^2,).
        Liouv (Optional[np.ndarray]): Liouvillian superoperator.
        DT (Optional[float]): Time step for evolution.
        TIME_STEPS (Optional[int]): Number of time steps.
        time_array (Optional[np.ndarray]): Array of time points.
        Gt (Optional[np.ndarray]): Time-evolution propagator.
    """
    
    def __init__(self, Nsys: int, Hsys: np.ndarray, rho0: np.ndarray) -> None:
        """
        Initialize the DynamicsGQME instance.
        
        Args:
            Nsys (int): System Hilbert space dimension.
            Hsys (np.ndarray): System Hamiltonian of shape (N, N).
            rho0 (np.ndarray): Initial density matrix of shape (N, N).
        """
        self.Nsys           =    Nsys
        self.Hsys           =    Hsys
        self.rho0           =    rho0
        self.vec_rho0       =    rho0.reshape(Nsys**2)
        
        #Liouvillian
        self.Liouv          =   None
        self.get_Liouvillian()    
        
        #time 
        self.DT             =   None
        self.TIME_STEPS     =   None
        self.time_array     =   None
        
        #propagator
        self.Gt             =   None

    
    def get_Liouvillian(self) -> None:
        """
        Construct the Liouvillian superoperator using the system Hamiltonian.
        
        The Liouvillian is defined as:
            L = H \otimes I - I \otimes H^T
        """
        Isys = np.eye(self.Nsys)
        self.Liouv = np.kron(self.Hsys,Isys) -  np.kron(Isys,self.Hsys.T)
        
    def setup_propagator(self, Gt: np.ndarray) -> None:
        """
        Set the time-evolution propagator.

        Args:
            Gt (np.ndarray): Time-dependent propagator.
        """
        self.Gt = Gt        
    
    def setup_timestep(self, DT: float, TIME_STEPS: int) -> None:
        """
        Set up the time discretization parameters.

        Args:
            DT (float): Time step size.
            TIME_STEPS (int): Number of time steps.
        """
        self.DT = DT
        self.TIME_STEPS = TIME_STEPS
        self.time_array = np.linspace(0,(TIME_STEPS-1)*DT,TIME_STEPS)

    def tt_tfd(
        self,
        initial_state: int = 0,
        update_type: str = 'rk4',
        rk4slices: int = 1,
        mmax: int = 4,
        RDO_arr_bench: Optional[np.ndarray] = None,
        show_steptime: bool = False
        ) -> Tuple[np.ndarray, np.ndarray]:
        
        """
        Perform Tensor-Train Thermofield Dynamics (TT-TFD) calculation.
        
        Args:
            initial_state (int, optional): Index of the initial state. Defaults to 0.
            update_type (str, optional): Method for time evolution. Either 'rk4' or 'krylov'.
            rk4slices (int, optional): Number of RK4 substeps if update_type is 'rk4'. Defaults to 1.
            mmax (int, optional): Dimension of Krylov subspace if update_type is 'krylov'. Defaults to 4.
            RDO_arr_bench (Optional[np.ndarray], optional): Benchmark RDO array to compute error at each step. Defaults to None.
            show_steptime (bool, optional): Whether to print timing for each step. Defaults to False.
        
        Returns:
            Tuple[np.ndarray, np.ndarray]: Time array and RDO (reduced density operator) array.
        """
        time_arr, RDO_arr = tfd.tt_tfd(
            initial_state,
            update_type,
            rk4slices=rk4slices,
            mmax=mmax,
            RDO_arr_bench=RDO_arr_bench,
            show_steptime=show_steptime
        )

        return time_arr, RDO_arr
    
    def cal_propagator_tttfd(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculate the numerical exact propagator for Spin-Boson model using the TT-TFD method.

        Returns:
            Tuple[np.ndarray, np.ndarray]: Time array and 3D propagator array.
        """
        
        U = np.zeros((pa.TIME_STEPS, pa.DOF_E_SQ, pa.DOF_E_SQ), dtype=np.complex128)
    
        # tt-tfd with initial state 0,1,2,3
        # initial state |0> means donor state |D>, |3> means acceptor state |A>
        # |1> is (|D> + |A>)/sqrt(2), |2> is (|D> + i|A>)/sqrt(2)
        for i in range(4):
            print(f"======== calculate the propagator, starting from state {i} ========")
            t, U[:, :, i] = tfd.tt_tfd(i)
   
        U_final = U.copy()
    
        # the coherence elements that start at initial state |D><A| and |A><D|
        # is the linear combination of above U results
        # |D><A| = |1><1| + i * |2><2| - 1/2 * (1 + i) * (|0><0| + |3><3|)
        U_final[:,:,1] = U[:,:,1] + 1.j * U[:,:,2] - 0.5 * (1. + 1.j) * (U[:,:,0] + U[:,:,3])
    
        # |A><D| = |1><1| - i * |2><2| - 1/2 * (1 - i) * (|0><0| + |3><3|)
        U_final[:,:,2] = U[:,:,1] - 1.j * U[:,:,2] - 0.5 * (1. - 1.j) * (U[:,:,0] + U[:,:,3])

        self.setup_propagator(U_final)
        print('========calculate the propagator done========')
        
        return t,U_final
    
    def prop_puresystem(self) -> np.ndarray:
        """
        Propagate the pure system under the unitary evolution.

        Returns:
            np.ndarray: Vectorized density matrix at each time step (shape (TIME_STEPS, N^2)).
        """
        Nstep = len(self.time_array)
        vec_rho = np.zeros((Nstep, self.Nsys**2), dtype=np.complex128)
        vec_rho[0] = self.vec_rho0.copy()
        for i in range(1,Nstep):
            vec_rho[i] = LA.expm(-1j*self.Liouv*(self.time_array[i]-self.time_array[i-1]))@vec_rho[i-1]
        return vec_rho
        
    def cal_F(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the first and second order time derivatives of the propagator.

        Returns:
            Tuple[np.ndarray, np.ndarray]: F and Fdot matrices (each of shape (TIME_STEPS, N^2, N^2)).
        """

        F = np.zeros((self.TIME_STEPS, self.Nsys**2, self.Nsys**2), dtype=np.complex128)
        Fdot = np.zeros((self.TIME_STEPS, self.Nsys**2, self.Nsys**2), dtype=np.complex128)
        
        if(self.Gt is None):
            print('error in cal_F: please setup the pre-calculated propagator super-operator')
            sys.exit()
            
        for j in range(self.Nsys**2):
            for k in range(self.Nsys**2):
                # extracts real and imag parts of U element
                Ureal = self.Gt[:,j,k].copy().real
                Uimag = self.Gt[:,j,k].copy().imag
    
                # F = i * d/dt U so Re[F] = -1 * d/dt Im[U] and Im[F] = d/dt Re[U]
                Freal = -1. * np.gradient(Uimag.flatten(), self.DT, edge_order = 2)
                Fimag = np.gradient(Ureal.flatten(), self.DT, edge_order = 2)
    
                # Fdot = d/dt F so Re[Fdot] = d/dt Re[F] and Im[Fdot] = d/dt Im[F]
                Fdotreal = np.gradient(Freal, self.DT)
                Fdotimag = np.gradient(Fimag, self.DT)
    
                F[:,j,k] = Freal[:] + 1.j * Fimag[:]
                Fdot[:,j,k] = Fdotreal[:] + 1.j * Fdotimag[:]
               
        return F,Fdot
    
        
    def _CalculateIntegral(
        self,
        F: np.ndarray,
        linearTerm: np.ndarray,
        prevKernel: np.ndarray,
        kernel: np.ndarray
        ) -> np.ndarray:
        """
        Compute the Volterra integral using the trapezoidal rule.

        Args:
            F (np.ndarray): Derivative of the propagator.
            linearTerm (np.ndarray): Linear term from memory kernel equation.
            prevKernel (np.ndarray): Kernel from previous iteration.
            kernel (np.ndarray): Kernel to be updated.

        Returns:
            np.ndarray: Updated kernel.
        """
        
        # time step loop starts at 1 because K is equal to linear part at t = 0
        for n in range(1, self.TIME_STEPS):
            kernel[n,:,:] = 0.
    
            # f(a) and f(b) terms
            kernel[n,:,:] += 0.5 * self.DT * F[n,:,:] @ kernel[0,:,:]
            kernel[n,:,:] += 0.5 * self.DT * F[0,:,:] @ prevKernel[n,:,:]
    
            # sum of f(a + kh) term
            for c in range(1, n):
                # since a new (supposed-to-be-better) guess for the
                # kernel has been calculated for previous time steps,
                # can use it rather than prevKernel
                kernel[n,:,:] += self.DT * F[n - c,:,:] @ kernel[c,:,:]
    
            # multiplies by i and adds the linear part
            kernel[n,:,:] = 1.j * kernel[n,:,:] + linearTerm[n,:,:]
    
        return kernel
    
    def get_memory_kernel(self) -> np.ndarray:
        """
        Compute the memory kernel using the Volterra scheme.

        Returns:
            np.ndarray: Memory kernel array (shape (TIME_STEPS, N^2, N^2)).
        """
        F,Fdot = self.cal_F()
        
        linearTerm = 1.j * Fdot.copy() # first term of the linear part
        for l in range(self.TIME_STEPS):
            # subtracts second term of linear part
            linearTerm[l,:,:] -= 1./pa.HBAR * F[l,:,:] @ self.Liouv
            
        START_TIME = time.time() # starts timing
        # sets initial guess to the linear part
        prevKernel = linearTerm.copy()
        kernel = linearTerm.copy()

        # loop for iterations
        for numIter in range(1, pa.MAX_ITERS + 1):
        
            iterStartTime = time.time() # starts timing of iteration
            print("Iteration:", numIter)
        
            # calculates kernel using prevKernel and trapezoidal rule
            kernel = self._CalculateIntegral(F, linearTerm, prevKernel, kernel)
        
            numConv = 0 # parameter used to check convergence of entire kernel
            for i in range(self.Nsys**2):
                for j in range(self.Nsys**2):
                    for n in range(self.TIME_STEPS):
                        # if matrix element and time step of kernel is converged, adds 1
                        if abs(kernel[n][i][j] - prevKernel[n][i][j]) <= pa.CONVERGENCE_PARAM:
                            numConv += 1
        
                        # if at max iters, prints which elements and time steps did not
                        # converge and prevKernel and kernel values
                        elif numIter == pa.MAX_ITERS:
                            print("\tK time step and matrix element that didn't converge: %s, %s%s"%(n,i,j))
        
            print("\tIteration time:", time.time() - iterStartTime)
        
            # enters if all times steps and matrix elements of kernel converged
            if numConv == self.TIME_STEPS * self.Nsys**2 * self.Nsys**2:
                # prints number of iterations and time necessary for convergence
                print("Number of Iterations:", numIter, "\tVolterra time:", time.time() - START_TIME)
        
                break # exits the iteration loop
        
            # if not converged, stores kernel as prevKernel, zeros the kernel, and then
            # sets kernel at t = 0 to linear part
            prevKernel = kernel.copy()
            kernel = linearTerm.copy()
        
            # if max iters reached, prints lack of convergence
            if numIter == pa.MAX_ITERS:
                print("\tERROR: Did not converge for %s iterations"%pa.MAX_ITERS)
                print("\tVolterra time:", print(time.time() - START_TIME))
        
        return kernel


    def PropagateRK4(
        self,
        currentTime: float,
        memTime: float,
        kernel: np.ndarray,
        sigma_hold: np.ndarray,
        sigma: np.ndarray
        ) -> np.ndarray:
        """
        Perform one 4th-order Runge-Kutta (RK4) integration step for GQME.
    
        Args:
            currentTime (float): Current time.
            memTime (float): Memory time cutoff.
            kernel (np.ndarray): Memory kernel array (shape (TIME_STEPS, N^2, N^2)).
            sigma_hold (np.ndarray): Current vectorized reduced density matrix (N^2,).
            sigma (np.ndarray): Array of vectorized reduced density matrix up to current time (TIME_STEPS, N^2).
    
        Returns:
            np.ndarray: Updated vectorized reduced density matrix after one RK4 step (N^2,).
        """
        f_0 = self._Calculatef(currentTime, memTime,
                         kernel, sigma, sigma_hold)
    
        k_1 = sigma_hold + self.DT * f_0 / 2.
        f_1 = self._Calculatef(currentTime + self.DT / 2., memTime,
                         kernel, sigma, k_1)
    
        k_2 = sigma_hold + self.DT * f_1 /2.
        f_2 = self._Calculatef(currentTime + self.DT / 2., memTime,
                         kernel, sigma, k_2)
    
        k_3 = sigma_hold + self.DT * f_2
        f_3 = self._Calculatef(currentTime + self.DT, memTime,
                         kernel, sigma, k_3)
    
        sigma_hold += self.DT / 6. * (f_0 + 2. * f_1 + 2. * f_2 + f_3)
    
        return sigma_hold
    
    def _Calculatef(
                    self,
                    currentTime: float,
                    memTime: float,
                    kernel: np.ndarray,
                    sigma_array: np.ndarray,
                    kVec: np.ndarray
                ) -> np.ndarray:
        """
        Evaluate the time derivative in GQME.
        
        Args:
            currentTime (float): Current time.
            memTime (float): Memory time cutoff.
            kernel (np.ndarray): Memory kernel (shape (TIME_STEPS, N^2, N^2)).
            sigma_array (np.ndarray): History of vectorized system density matrix (TIME_STEPS, N^2).
            kVec (np.ndarray): Vector to be evolved (N^2,).
        
        Returns:
            np.ndarray: Time derivative vector f(t) (N^2,).
        """
        memTimeSteps = int(memTime / self.DT)
        currentTimeStep = int(currentTime / self.DT)
    
        f_t = np.zeros(kVec.shape, dtype=np.complex128)
    
        f_t -= 1.j / pa.HBAR * self.Liouv @ kVec
    
        limit = memTimeSteps
        if currentTimeStep < (memTimeSteps - 1):
            limit = currentTimeStep
        for l in range(limit):
            f_t -= self.DT * kernel[l,:,:] @ sigma_array[currentTimeStep - l]
    
        return f_t


    def solve_gqme(
                self,
                kernel: np.ndarray,
                MEM_TIME: float,
                dtype: str = "Density"
            ) -> np.ndarray:
        """
        Solve the GQME using RK4 integration.
    
        Args:
            kernel (np.ndarray): Memory kernel (shape (TIME_STEPS, N^2, N^2)).
            MEM_TIME (float): Memory cutoff time.
            dtype (str): Type of data to propagate: "Density" or "Propagator".
    
        Returns:
            np.ndarray: Propagated state (shape depends on dtype).
        """
        
        if(dtype=="Density"):
            # array for reduced density matrix elements
            sigma = np.zeros((self.TIME_STEPS, self.Nsys**2), dtype=np.complex128)
            # array to hold copy of sigma
            sigma_hold = np.zeros(self.Nsys**2, dtype = np.complex128)
            
            # sets the initial state
            sigma[0,:] = self.vec_rho0.copy()
            sigma_hold = self.vec_rho0.copy()
        elif(dtype == "Propagator"):
            # array for reduced density matrix elements
            sigma = np.zeros((self.TIME_STEPS, self.Nsys**2, self.Nsys**2), dtype=np.complex128)
            # array to hold copy of sigma
            sigma_hold = np.zeros(self.Nsys**2, dtype = np.complex128)
            
            #time 0 propagator: identity superoperator
            sigma[0] = np.eye(self.Nsys**2)
            #array to hold copy of G propagator
            sigma_hold = np.eye((self.Nsys**2), dtype=np.complex128)
        else:
            sys.exit('GQME input error, dtype should be "Density" or "Propagator"')

        # loop to propagate sigma
        print(">>> Starting GQME propagation, memory time =", MEM_TIME)
        for l in range(self.TIME_STEPS - 1): # it propagates to the final time step
            if l%100==0: print(l)
            currentTime = l * self.DT
        
            sigma_hold = self.PropagateRK4(currentTime, MEM_TIME, kernel, sigma_hold, sigma)
        
            sigma[l + 1] = sigma_hold.copy()
        
        return sigma



import numpy as np
from itertools import combinations, product
from scipy.linalg import expm, kron
from numpy.random import rand
random_numbers = np.random.uniform(low=10000, high=99999)
np.random.seed(int(random_numbers))
from .ansatz import *
from .ansatzVect import *

# Helper Functions for evolution
# def tag(A):
#     return A.tag if A else None #!!!!!!!a.any or a.all??
def tag(A):
    return getattr(A, 'tag', None) if A is not None else None


def aexp(A, theta):
    return expm(-1j * theta * A.mat / 2)

def add_A(A, P):
    A.A.append(P)
    A.theta = np.append(A.theta, 0)

def lmul(A, ψ):
    # print("ehrerw:",A)
    # print("amat:",np.shape(A.mat))
    # print("her in lmul",A.mat @ ψ)
    return A.mat @ ψ

def update_theta(ansatz, dtheta, dt):
    ansatz.theta = ansatz.theta + dtheta * dt

def partial_theta(A):
    len_A = len(A.A)
    ψ = A.ref.copy()
    res = []
    for i in range(len_A):
        ψ = aexp(A.A[i], A.theta[i]) @ ψ
        ψt = -0.5j * lmul(A.A[i], ψ)
        for j in range(i + 1, len_A):
            ψt = aexp(A.A[j], A.theta[j]) @ ψt
        res.append(ψt)
    return res

def build_m(ψ, dψ):
    l = len(dψ)
    res = np.zeros((l, l))
    ψ = ψ[:, np.newaxis] if len(ψ.shape) == 1 else ψ

    for μ in range(l):
        for ν in range(l):
            res[μ, ν] = 2 * np.real(dψ[μ].T.conj() @ dψ[ν] + ψ.T.conj() @ dψ[μ] @ ψ.T.conj() @ dψ[ν])
    return res

def update_m(m, dψa, ψ, dψ):
    l = m.shape[0]
    mp = np.zeros((l + 1, l + 1))
    dψa = dψa[:, np.newaxis]
    ψ = ψ[:, np.newaxis]
    mp[l, l] = 2 * np.real((dψa.T.conj() @ dψa + ψ.T.conj() @ dψa @ ψ.T.conj() @ dψa).item())
    mp[:l, :l] = m
    for μ in range(l):
        temp = 2 * np.real(dψ[μ].T.conj() @ dψa + ψ.T.conj() @ dψ[μ] @ ψ.T.conj() @ dψa)
        mp[μ, l] = temp.item()
        mp[l, μ] = temp.item()
    return mp

def build_v(ψ, dψ, He, Ha):
    L = -1j * (He - 1j * Ha)
    l = len(dψ)
    ψ = ψ[:, np.newaxis] if len(ψ.shape) == 1 else ψ
    res = np.zeros(l)
    for μ in range(l):
        res[μ] = 2 * np.real((dψ[μ].T.conj() @ L @ ψ + dψ[μ].T.conj() @ ψ @ ψ.T.conj() @ L.T.conj() @ ψ).item())
    return res

def update_v(v, dψₐ, ψ, He, Ha):
    L = -1j * (He - 1j * Ha)
    l = len(v)
    dψₐ = dψₐ[:, np.newaxis]
    ψ = ψ[:, np.newaxis]
    vp = np.zeros(l + 1)
    vp[:l] = v
    vp[l] = 2 * np.real((dψₐ.T.conj() @ L @ ψ + dψₐ.T.conj() @ ψ @ ψ.T.conj() @ L.T.conj() @ ψ).item())
    return vp

def lin_solve(m, v, λ=1e-4):
    A = m.T @ m
    try:
        xλ = np.linalg.solve(A + λ**2 * np.eye(A.shape[0]), m.T @ v)
    except np.linalg.LinAlgError:
        xλ = np.linalg.lstsq(A + λ**2 * np.eye(A.shape[0]), m.T @ v, rcond=None)[0]
    return xλ, 2 * v.T @ xλ - xλ.T @ m @ xλ

def get_newest_A(A):
    return A.A[-1] if A.A else None

def update_state(A):
    if A.A:
        state = A.ref.copy()
        for i in range(len(A.A)):
            state = aexp(A.A[i], A.theta[i]) @ state
        A.state = state
    else:
        A.state = A.ref

def set_ref(A, ref):
    A.ref = ref

def reset(A):
    A.theta = np.array([])
    A.A = []
    update_state(A)

class Result:
    def __init__(self, t, psi, energy, pop, theta, A, jump_t, jump_L, status):
        self.t = t
        self.psi = psi
        self.energy = energy  # Add energy attribute
        self.pop = pop
        self.theta = theta
        self.A = A
        self.jump_t = jump_t
        self.jump_L = jump_L
        self.status = status

    def __repr__(self):
        return f"Result(t={self.t}, psi={self.psi}, energy={self.energy}, pop={self.pop}, theta={self.theta}, A={self.A}, jump_t={self.jump_t}, jump_L={self.jump_L}, status={self.status})"

class AVQDSol:
    def __init__(self, t, u, θ, A, jump_L, jump_t, status=0, norm=[]):
        self.t = t
        self.u = u
        self.θ = θ
        self.A = A
        self.jump_L = jump_L
        self.jump_t = jump_t
        self.status = status
        self.norm = norm

def one_step(A, He, Ha, dt):

    relrcut = A.relrcut
    psi_ = A.state
    dpsi_ = partial_theta(A)
    M = build_m(psi_, dpsi_)
    V = build_v(psi_, dpsi_, He, Ha)
    dtheta, vmv = lin_solve(M, V)
    vmvMax = vmv
    Mtmp = M
    Vtmp = V
    dthetatmp = dtheta
    opTmp = None
    dpsi_ₐTmp = None
    add_flag = True

    while add_flag:
        # print("here in onestep While")
        tagTmp = None
        # print("Ansatz pool:",A.pool)
        for op in A.pool:
            # print("opertaro from pool:",(op.tag))
            if tag(get_newest_A(A)) == tag(op):
                # print("here in get newest_A")
                continue
            # print("here in op for loop")
            dpsi_ₐ = -0.5j * lmul(op, psi_)

            Mop = update_m(M, dpsi_ₐ, psi_, dpsi_)
            # print("Mop:",Mop)
            Vop = update_v(V, dpsi_ₐ, psi_, He, Ha)
            dthetaop, vmvOp = lin_solve(Mop, Vop)
            # print("berfore the if:","vmvOp:",vmvOp, "vmvMax:",vmvMax)

            if vmvOp > vmvMax:
                # print("vmvOp:",vmvOp, "vmvMax:",vmvMax)

                Mtmp = Mop
                Vtmp = Vop
                vmvMax = vmvOp
                dthetatmp = dthetaop
                opTmp = op
                tagTmp = tag(op)
                # print("tagTmp:",tagTmp)
                dpsi_ₐTmp = dpsi_ₐ
        # print("vmvMax - vmv:",vmvMax-vmv)
        add_flag = vmvMax - vmv >= relrcut

        

        if tagTmp is not None and add_flag:

            add_A(A, opTmp)
            vmv = vmvMax
            M = Mtmp
            V = Vtmp
            dtheta = dthetatmp
            dpsi_.append(dpsi_ₐTmp)

    update_theta(A, dtheta, dt)
    # print("Ansatz:",A.theta)
    update_state(A)


def solve_avq_traj(H, A, tspan, dt, save_state=True, save_everystep=True):
    # Store initial reference state for reinitialization
    ref_init = A.ref.copy()
    update_state(A)
    
    # Initialize time and recorder lists
    t = tspan[0]
    t_list = []
    u_list = []
    theta_list = []
    A_list = []
    jump_L_list = []
    jump_t_list = []

    # Break flag for error handling
    break_flag = False
    Gamma = 0  # Jump recorder
    q = rand()  # Random number for quantum jump
    # print("expo:",np.exp(-Gamma))
    # print("random q:", q)

    if save_everystep:
        t_list.append(t)
        theta_list.append(A.theta.copy())

        A_list.append([tag(a) for a in A.A])
        if save_state:
            u_list.append(A.state.copy())

    # Main evolution loop
    # This loop performs the simulation of quantum system evolution over a set time span.
    # It integrates the system state forward in time while handling both deterministic 
    # evolution and stochastic quantum jumps.

    while t + dt <= tspan[1]:  # Loop over the total time span with step size dt
        # print("timestep:", dt)  # Print current time step size for monitoring
        He = H.He  # Extract Hermitian Hamiltonian component for deterministic evolution
        Ha = H.Ha  # Extract Antihermitian Hamiltonian component for jump handling

        if np.exp(-Gamma) > q:  # Check if no quantum jump occurs, based on jump probability threshold
            try:
                one_step(A, He, Ha, dt)  # Perform deterministic evolution for one time step
            except Exception:  # In case of failure, exit the loop with a flag
                break_flag = True
                break
            psi_ = A.state  # Update the current state of the system
            Gamma += 2 * np.real(psi_.T.conj() @ Ha @ psi_) * dt  # Update accumulated Gamma for jump probability

            if save_everystep:  # Option to save state at every time step
                t_list.append(t + dt)  # Save current time
                theta_list.append(A.theta.copy())  # Save current ansatz parameters
                A_list.append([tag(a) for a in A.A])  # Save current state of ansatz components

        else:  # Quantum jump occurs
            psi_ = A.state  # Get current state before jump

            # Compute the probabilities for different jump operators
            weights = np.array([np.real(psi_.T.conj() @ LdL @ psi_) for LdL in H.LdL])

            # Randomly choose a jump operator based on computed probabilities
            L_index = np.random.choice([i for i in range(np.shape(H.Llist)[0])], p=weights/weights.sum())
            L = H.Llist[L_index]  # Select corresponding jump operator

            # Apply the jump operator to the state and normalize it
            psi_n = L @ psi_ / np.linalg.norm(L @ psi_)

            # Record the jump event details (time and operator used)
            jump_t_list.append(t)
            jump_L_list.append(tag(L))

            # Record the state and parameters post-jump
            t_list.append(t + dt)
            set_ref(A, psi_n)  # Update the state with post-jump state
            reset(A)  # Reset the ansatz parameters after jump

            # Save updated ansatz parameters post-jump
            theta_list.append(A.theta.copy())
            A_list.append([tag(a) for a in A.A])

            # Reset the jump probability accumulator
            Gamma = 0
            q = rand()  # Generate a new random number for the next jump check

        # Optionally save the state at each time step
        if save_state:
            u_list.append(A.state.copy())  # Store the current state

        # Increment the simulation time by dt
        t += dt


    # Final check for status if no errors occurred
    if not break_flag:
        if (not jump_t_list or not np.isclose(t, jump_t_list[-1])) and not save_everystep:
            theta_list.append(A.theta.copy())
            A_list.append([tag(a) for a in A.A])

    # Reset the ansatz to the initial condition
    # print("ansatz before reset:",A.A)
    set_ref(A, ref_init)
    reset(A)
    # print("tlist:",t_list)
    # Return solution object
    return AVQDSol(t_list, u_list, theta_list, A_list, jump_L_list, jump_t_list, status=0 if not break_flag else 1)

# Function to evolve a single trajectory and collect results
def solve_avq_trajectory(H, ansatz, tf, dt):
    # Solve for a single trajectory using solve_avq
    res = solve_avq_traj(H, ansatz, [0, tf], dt, save_state=True, save_everystep=True)
    
    # Post-process the results to extract energy and populations
    tlist = res.t
    psi_list = res.u
    energy = []
    pop = []
    
    for i in range(len(tlist)):
        psi = psi_list[i]
        energy.append(np.real(np.conj(psi.T) @ (H.He) @ psi))  # Energy
        pop.append([np.abs(psi[0])**2, np.abs(psi[1])**2])  # Population in ground/excited states for amplitude damping

    return Result(tlist, psi_list, energy, pop, res.θ, res.A, res.jump_t, res.jump_L, res.status)
    pass

def solve_avq_vect(H, A, tspan, dt):
    ref_init = A.ref.copy()
    nqbit = A.nqbit
    update_state(A)
    t = tspan[0]
    t_list = [t]
    u_list = [A.state.reshape(2**nqbit, 2**nqbit)]
    theta_list = [A.theta.copy()]
    A_list = [[tag(a) for a in A.A]]
    norm_list = [1.0]
    Gamma = 0
    while t + dt <= tspan[1]:
        He = H.He
        Ha = H.Ha
        one_step(A, He, Ha, dt)
        psi_ = A.state
        Gamma += 2 * np.real(psi_.T.conj() @ Ha @ psi_) * dt
        ρ = psi_.reshape(2**nqbit, 2**nqbit)
        ρ /= np.trace(ρ)
        t += dt
        t_list.append(t)
        u_list.append(ρ)
        theta_list.append(A.theta.copy())
        A_list.append([tag(a) for a in A.A])
        norm_list.append(np.exp(-Gamma))
    set_ref(A, ref_init)
    reset(A)
    return AVQDSol(t_list, u_list, theta_list, A_list, [], [], norm=norm_list)



# Utility Functions/Patches:
from qiskit import transpile


def execute(QCircuit, backend=None, shots=None):
    '''
        Function to replace the now-deprecated Qiskit
        `QuantumCircuit.execute()` method.

        Input:
          - `QCircuit`: qiskit.QuantumCircuit object
          - `Backend`: qiskit.Backend instance
          - `shots`: int specifying the number of shots
    '''
    # Transpile circuit with statevector backend
    tmp_circuit = transpile(QCircuit, backend)
    # Run the transpiled circuit
    if shots:
        job = backend.run(tmp_circuit, n_shots=shots)
    else:
        job = backend.run(tmp_circuit)
    return(job)



import numpy as np
from itertools import combinations, product
from scipy.linalg import expm, kron

# Ansatz Class and Operators
class Ansatz_class:
    def __init__(self, nqbit, u0, relrcut, pool, theta=[], ansatz=[]):
        self.theta = np.array(theta)
        self.A = ansatz
        self.state = u0.copy()
        self.ref = u0.copy()
        self.relrcut = relrcut
        self.nqbit = nqbit
        self.pool = pool

class AnsatzOperatorBase:
    pass

class PauliOperator_class(AnsatzOperatorBase):
    def __init__(self, mat, tag, nqbit):
        self.mat = mat
        self.tag = tag
        self.nqbit = nqbit

def single_clause(ops, q_ind, weight, num_qubit):
    si = np.eye(2)
    res = weight * np.eye(1)
    for i in range(1, num_qubit + 1):
        if i in q_ind:
            op2 = eval(f"{ops[q_ind.index(i)]}")
            res = kron(res, op2)
        else:
            res = kron(res, si)
    return res

def PauliOperator(ops, idx, w, nqbit):
    sortedRep = sorted(zip(ops, idx), key=lambda x: x[1])
    tag = ''.join([f"{o}{i}" for o, i in sortedRep])
    mat = single_clause(ops, idx, w, nqbit)
    return PauliOperator_class(mat, tag, nqbit)

# Defining pool of operators/gates
sx = np.array([[0, 1], [1, 0]])
sy = np.array([[0, -1j], [1j, 0]])
sz = np.array([[1, 0], [0, -1]])

def build_pool(nqbit):
    pauliStr = ["sx", "sz", "sy"]
    res = []
    for order in range(1, nqbit+1):
        for idx in combinations(range(1, nqbit + 1), order):
            for op in product(pauliStr, repeat=order):
                res.append(PauliOperator(op, list(idx), 1, nqbit))
    return res

def Ansatz(u0, relrcut, theta=[], ansatz=[]):
    nqbit = int(np.log2(len(u0)))
    pool_qubit = 2 * nqbit
    u0 = np.outer(u0, u0).flatten()
    pool = build_pool(pool_qubit)
    return Ansatz_class(nqbit, u0, relrcut, pool, theta, ansatz)
